import os
from typing import Tuple

import numpy as np
import pandas as pd

from beetroots.inversion.plots.map_shaper import MapShaper
from beetroots.simulations.astro import data_validation
from beetroots.simulations.astro.real_data.real_data_nn import SimulationRealDataNN


def read_point_challenger(
    path_data_cloud: str,
    filename: str,
    point_name: str,
) -> dict:
    df_pt_challenger = pd.read_pickle(f"{path_data_cloud}/{filename}")
    point_challenger = {"name": point_name, "value": df_pt_challenger.values}
    return point_challenger


def apply_data_transformation_3cases_rule(
    df_int: pd.DataFrame,
    df_err: pd.DataFrame,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    df_int_inter = df_int.drop("idx", 1)
    df_err_inter = df_err.drop("idx", 1)
    df_censor_inter = df_err_inter.copy()

    # case 1 : sigma = + infty : omega = u and sigma = 1/3 * u
    mask = np.isinf(df_err)
    df_err_inter[mask] = 1 / 3 * df_int[mask]
    df_censor_inter[mask] = 1 * df_int[mask]

    # case 2 : u < sigma < + infty : omega = 3 * sigma and u = omega
    mask = (df_int < df_err) & ~np.isinf(df_err) & ~np.isnan(df_err)
    df_censor_inter[mask] = 3 * df_err[mask]
    df_int_inter[mask] = 1 * df_censor_inter[mask]

    # case 3 : sigma < u < + infty : omega = sigma
    mask = (df_err < df_int) & ~np.isinf(df_err) & ~np.isnan(df_err)
    df_censor_inter[mask] = 1 * df_err[mask]

    return df_int_inter, df_err_inter, df_censor_inter


def apply_data_transformation_3cases_rule_alternative(
    df_int: pd.DataFrame,
    df_err: pd.DataFrame,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    df_int_inter = df_int.drop("idx", 1)
    df_err_inter = df_err.drop("idx", 1)
    df_censor_inter = df_err_inter.copy()

    # first : never consider observations to be censored
    df_censor_inter = 1e-60

    # if sigma = + infty : put very high noise std
    mask = np.isinf(df_err)
    df_err_inter[mask] = 1.0

    return df_int_inter, df_err_inter, df_censor_inter


class SimulationAstroCarina(SimulationRealDataNN):
    def setup_observation(
        self,
        data_int_path: str,
        data_err_path: str,
    ) -> Tuple[pd.DataFrame, np.ndarray, np.ndarray, np.ndarray]:
        df_int = pd.read_pickle(data_int_path)
        df_err = pd.read_pickle(data_err_path)
        assert list(df_int.index.names) == ["X", "Y"]
        assert list(df_err.index.names) == ["X", "Y"]
        df_int = df_int.sort_index()
        df_err = df_err.sort_index()

        if "idx" not in list(df_err.columns):
            df_int["idx"] = np.arange(len(df_int))
            df_err["idx"] = np.arange(len(df_err))

        assert df_int.shape == df_err.shape
        assert list(df_int.columns) == list(df_err.columns)

        for line in self.list_lines_fit:
            assert line in list(df_int.columns)

        self.list_lines_valid = list(
            set(list(df_int.columns)) - set(self.list_lines_fit)
        )
        self.list_lines_valid.remove("idx")

        # * select lines
        df_int_fit = df_int.loc[:, self.list_lines_fit + ["idx"]]
        df_err_fit = df_err.loc[:, self.list_lines_fit + ["idx"]]

        df_int_valid = df_int.loc[:, self.list_lines_valid + ["idx"]]
        df_err_valid = df_err.loc[:, self.list_lines_valid + ["idx"]]

        # eliminate absurd values
        for col in self.list_lines_fit:
            df_int_fit[col] = df_int_fit[col].apply(
                lambda x: x if x <= 1e3 else np.inf,
            )
            df_err_fit[col] = df_err_fit[col].apply(
                lambda x: x if x <= 1e3 else np.inf,
            )
        for col in self.list_lines_valid:
            df_int_valid[col] = df_int_valid[col].apply(
                lambda x: x if x <= 1e3 else np.inf,
            )
            df_err_valid[col] = df_err_valid[col].apply(
                lambda x: x if x <= 1e3 else np.inf,
            )

        self.N = len(df_int)

        # * correct values (censoring / errors / etc.)
        df_censor_fit = df_err_fit.copy()
        df_censor_fit.loc[:, self.list_lines_fit] *= 3  # omega = 3 sigma_a

        df_censor_valid = df_err_valid.copy()
        df_censor_valid.loc[:, self.list_lines_valid] *= 3  # omega = 3 sigma_a

        # * apply data transformation 3-cases rule
        (
            df_int_fit_inter,
            df_err_fit_inter,
            df_censor_fit_inter,
        ) = apply_data_transformation_3cases_rule_alternative(
            df_int_fit,
            df_err_fit,
        )
        (
            df_int_valid_inter,
            df_err_valid_inter,
            df_censor_valid_inter,
        ) = apply_data_transformation_3cases_rule_alternative(
            df_int_valid,
            df_err_valid,
        )

        df_int_fit.loc[:, self.list_lines_fit] = df_int_fit_inter
        df_err_fit.loc[:, self.list_lines_fit] = df_err_fit_inter
        df_censor_fit.loc[:, self.list_lines_fit] = df_censor_fit_inter

        df_int_valid.loc[:, self.list_lines_valid] = df_int_valid_inter
        df_err_valid.loc[:, self.list_lines_valid] = df_err_valid_inter
        df_censor_valid.loc[:, self.list_lines_valid] = df_censor_valid_inter

        # *
        y_fit = np.nan_to_num(df_int_fit.drop("idx", 1).values)
        sigma_a_fit = np.nan_to_num(df_err_fit.drop("idx", 1).values, nan=1)
        omega_fit = df_censor_fit.drop("idx", 1).values

        y_valid = np.nan_to_num(df_int_valid.drop("idx", 1).values)
        sigma_a_valid = np.nan_to_num(df_err_valid.drop("idx", 1).values, nan=1)
        omega_valid = df_censor_valid.drop("idx", 1).values

        self.X_true_scaled = None
        self.map_shaper = MapShaper(df_int)

        # * save observation
        df_int_fit.to_pickle(f"{self.path_data_csv_in}/observation_maps.pkl")
        df_err_fit.to_pickle(f"{self.path_data_csv_in}/additive_std.pkl")
        df_censor_fit.to_pickle(
            f"{self.path_data_csv_in}/censor_threshold.pkl",
        )

        return (
            df_int_fit,
            y_fit,
            sigma_a_fit,
            omega_fit,
            y_valid,
            sigma_a_valid,
            omega_valid,
        )


if __name__ == "__main__":
    # * setup path to data
    path_data_cloud = f"{os.path.dirname(os.path.abspath(__file__))}"
    path_data_cloud += "/../../../../../data/carina"

    # * import Wu et al. 2018 values
    point_challenger = read_point_challenger(
        path_data_cloud,
        filename="values_wu_et_al_2018.pkl",
        point_name="Wu et al., 2018",
    )

    params = SimulationRealDataNN.load_params(path_data_cloud)
    SimulationRealDataNN.check_input_params_file(
        params,
        data_validation.schema,
    )

    simulation = SimulationRealDataNN(**params["simu_init"], params=params)

    simulation.main(
        params=params,
        path_data_cloud=path_data_cloud,
        point_challenger=point_challenger,
    )



import abc


class ApproxOptimForwardMap(abc.ABC):
    @abc.abstractmethod
    def compute_log10_f_Theta(self):
        r"""Evaluates $\log_{10} f$ on a list of `self.N_samples_theta` $x$ values, each randomly generated in the hypercube defined by its lower and upper bounds.

        Parameters
        ----------
        forward_model_name : str
            name of the forward model to load (i.e., of the corresponding folder)
        angle : float
            angle at which the cloud is observed
        lower_bounds_lin : np.ndarray of shape (D,)
            array of lower bounds in linear scale
        upper_bounds_lin : np.ndarray of shape (D,)
            array of upper bounds in linear scale

        Returns
        -------
        np.ndarray of shape (self.N_samples_theta, L)
            evaluations of the `self.N_samples_theta` $x$ values for the L considered lines.
        """
        pass



r"""contains the validation schema that ensures that the ``.yaml`` input file, that depicts the inversion configuration, has a correct structure.
"""

schema = {
    "simu_init": {
        "required": True,
        "type": "dict",
        "schema": {
            "simu_name": {"required": True, "type": "string"},
            "cloud_name": {"required": True, "type": "string"},
            "max_workers": {"required": True, "type": "number", "min": 1, "max": 20},
            # "params_names": {},
        },
    },
    #
    "to_run_optim_map": {"required": True, "type": "boolean"},
    "to_run_mcmc": {"required": True, "type": "boolean"},
    #
    # spatial prior
    "with_spatial_prior": {"required": True, "type": "boolean"},
    "spatial_prior": {
        "required": False,
        "type": "dict",
        "schema": {
            "name": {"type": "string", "allowed": ["L2-laplacian", "L2-gradient"]},
            "use_next_nearest_neighbors": {"type": "boolean"},
            "initial_regu_weights": {"type": "list"},
        },
    },
    # sampling params
}
r"""validation schema that ensures that the ``.yaml`` input file, that depicts the inversion configuration, has a correct structure"""



from typing import Tuple

import numba as nb
import numpy as np
from scipy.special import gamma, ndtr

from beetroots.modelling.priors import smooth_indicator_prior

GAMMA_1_4 = gamma(1 / 4)


@nb.njit()
def sample_generalized_gaussian(
    alpha: float,
    a_alpha: float,
    size: Tuple[int, int],
    seed: int,
) -> np.ndarray:
    r"""sample from a generalized Gaussian distribution of pdf:

    .. math::
        :label: eq:pdf_generalized_gaussian

        ..math::
        p(\theta) = \frac{2}{\delta \Gamma(1/4)} \exp \left\{- \left( \frac{\theta}{\delta}\right)^4\right\}

    with here :math:`\delta = 1/A(\alpha)`.
    See :cite:t:`nardonSimulationTechniquesGeneralized2009` for more details.

    Parameters
    ----------
    alpha : float
        power used in the distribution. alpha=2 is equivalent to the Gaussian distribution.
    a_alpha : float
        scaling parameter
    size : Tuple[int, int]
        number of samples to draw
    seed : int
        random seed for sampling reproducibility

    Returns
    -------
    np.ndarray
        array of samples
    """
    np.random.seed(seed)
    z = np.random.gamma(shape=1 / alpha, scale=(1 / a_alpha) ** alpha, size=size)
    x = np.random.binomial(n=1, p=0.5, size=size) * 2 - 1
    return x * z ** (1 / alpha)


@nb.njit()
def sample_smooth_indicator(
    lower_bounds: np.ndarray,
    upper_bounds: np.ndarray,
    indicator_margin_scale: float,
    size: Tuple[int, int],
    seed: int,
) -> np.ndarray:
    r"""algorithm to draw iid samples from the smooth indicator prior, detailed in Appendix B in :cite:t:`paludEfficientSamplingNon2023`

    Parameters
    ----------
    lower_bounds : np.ndarray
        vector of lower bounds for each of the D physical parameters
    upper_bounds : np.ndarray
        vector of upper bounds for each of the D physical parameters
    indicator_margin_scale : float
        scaling parameter, quantifies how much values out of the validity intervals are penalized
    size : Tuple[int, int]
        number of samples to draw
    seed : int
        random seed, for sampling reproducibility

    Returns
    -------
    np.ndarray
        iid samples drawn from the smooth uniform distribution
    """
    np.random.seed(seed)
    k_mtm, D = size

    Z_unif = 2 * (upper_bounds - lower_bounds) / (indicator_margin_scale * GAMMA_1_4)
    Z_tot = 1 + Z_unif
    w_unif = Z_unif / Z_tot  # (D,)

    z1 = np.zeros((k_mtm, D))
    x_unif = np.zeros((k_mtm, D))
    for d in range(D):
        z1[:, d] = np.random.binomial(1, w_unif[d], size=k_mtm)
        x_unif[:, d] = np.random.uniform(lower_bounds[d], upper_bounds[d], size=k_mtm)

    x_gg = sample_generalized_gaussian(4.0, 1 / indicator_margin_scale, size, seed)
    x_gg = np.where(
        x_gg < 0,
        x_gg + np.expand_dims(lower_bounds, 0),
        x_gg + np.expand_dims(upper_bounds, 0),
    )
    return np.where(z1 == 1, x_unif, x_gg)


@nb.njit()
def get_neighboring_pixels(
    current_Theta: np.ndarray, list_edges: np.ndarray, idx_pix: int
) -> np.ndarray:
    r"""returns the array of current values of the neighboring pixels of a specified pixel

    Parameters
    ----------
    current_Theta : np.ndarray
        current iterate
    list_edges : np.ndarray
        list of edges, defined as pairs of pixel indices. Generated by the class:`SpatialPrior`.
    idx_pix : int
        index of a pixel

    Returns
    -------
    np.ndarray
        array of current values of the neighboring pixels
    """
    list_edges_n = list_edges[
        (list_edges[:, 0] == idx_pix) | (list_edges[:, 1] == idx_pix)
    ]
    list_neighbors_idx = list_edges_n.flatten()
    list_neighbors_idx = list_neighbors_idx[list_neighbors_idx != idx_pix]
    return current_Theta[list_neighbors_idx, :]  # (n_neighbors, D)


@nb.njit()
def sample_conditional_spatial_and_indicator_prior(
    current_Theta: np.ndarray,
    spatial_list_edges: np.ndarray,
    spatial_weights: np.ndarray,
    indicator_lower_bounds: np.ndarray,
    indicator_upper_bounds: np.ndarray,
    indicator_indicator_margin_scale: float,
    idx_pix: np.ndarray,
    k_mtm: int,
    seed: int,
) -> np.ndarray:
    r"""draw ``k_mtm`` iid samples for each candidate in the `ìdx_pix`` array.

    Parameters
    ----------
    current_Theta : np.ndarray
        current iterate
    spatial_list_edges : np.ndarray
        list of edges, defined as pairs of pixel indices. Generated by the class:`SpatialPrior`.
    spatial_weights : np.ndarray
        regularization weights of the spatial prior
    indicator_lower_bounds : np.ndarray
        vector of lower bounds for the D physical parameters
    indicator_upper_bounds : np.ndarray
        vector of upper bounds for the D physical parameters
    indicator_indicator_margin_scale : float
        scaling parameter, quantifies how much values out of the validity intervals are penalized
    idx_pix : np.ndarray
        array of indices of pixels for which candidates are to be proposed
    k_mtm : int
        number of candidates in the MTM kernel
    seed : int
        random seed, used for sampling reproducibility

    Returns
    -------
    np.ndarray
        samples drawn from the conditional prior
    """
    np.random.seed(seed)
    (N, D) = current_Theta.shape
    n_pix = idx_pix.size * 1

    samples = np.zeros((n_pix, k_mtm, D))
    i = 0
    for idx_1_pix in idx_pix:
        # * sample from around neighbors
        neighbors = get_neighboring_pixels(current_Theta, spatial_list_edges, idx_1_pix)
        N_neighbors = neighbors.shape[0]

        if N_neighbors > 0:
            Theta = np.zeros((k_mtm, D))

            for k in range(k_mtm):
                # select which combination f neighbors is going to be used
                arr_use_neighbors = np.random.binomial(n=1, p=0.5, size=N_neighbors)
                while np.max(arr_use_neighbors) == 0:
                    arr_use_neighbors = np.random.binomial(n=1, p=0.5, size=N_neighbors)

                used_neighbors = neighbors[arr_use_neighbors == 1]
                N_used_neighbors = used_neighbors.shape[0]

                # sigma_mtm_eff = 1 / (2 * np.sqrt(N_neighbors * spatial_weights))  # (D,)
                sigma_mtm_eff = 1 / (
                    2 * np.sqrt(N_used_neighbors * spatial_weights)
                )  # (D,)

                # initialize array of candidates

                mean = np.zeros((1, D))
                for d in range(D):
                    mean[0, d] = np.mean(used_neighbors[:, d])
                    # mean[d] = np.mean(neighbors[:, d])

                repeat = True
                n_repeats = 0
                n_repeats_tot = 0
                while repeat:
                    # * step 1 : generate candidates from spatial prior only
                    Theta_cand = mean + sigma_mtm_eff * np.random.standard_normal(
                        size=(1, D)
                    )

                    # * step 2 : accept or reject with combination of spatial
                    # * and indicator priors
                    p_Theta = np.exp(
                        -smooth_indicator_prior.penalty_one_pix(
                            Theta_cand,
                            indicator_lower_bounds,
                            indicator_upper_bounds,
                            indicator_indicator_margin_scale,
                        )
                    )  # (1,)
                    u = np.random.uniform(0, 1)
                    if u <= p_Theta[0]:
                        Theta[k] += Theta_cand[0]
                        repeat = False

                    n_repeats += 1
                    n_repeats_tot += 1

                    if n_repeats >= 10:
                        # select which combination f neighbors is going to be used
                        arr_use_neighbors = np.random.binomial(
                            n=1, p=0.5, size=N_neighbors
                        )
                        while np.max(arr_use_neighbors) == 0:
                            arr_use_neighbors = np.random.binomial(
                                n=1, p=0.5, size=N_neighbors
                            )

                        used_neighbors = neighbors[arr_use_neighbors == 1]
                        N_used_neighbors = used_neighbors.shape[0]

                        # sigma_mtm_eff = 1 / (2 * np.sqrt(N_neighbors * spatial_weights))  # (D,)
                        sigma_mtm_eff = 1 / (
                            2 * np.sqrt(N_used_neighbors * spatial_weights)
                        )  # (D,)

                        # initialize array of candidates

                        mean = np.zeros((1, D))
                        for d in range(D):
                            mean[0, d] = np.mean(used_neighbors[:, d])
                            # mean[d] = np.mean(neighbors[:, d])

                        n_repeats = 0

                    assert n_repeats_tot < 1_000

            samples[i, :, :] = Theta * 1  # (k_mtm, D)

        else:
            samples[i, :, :] = sample_smooth_indicator(
                indicator_lower_bounds,
                indicator_upper_bounds,
                indicator_indicator_margin_scale,
                size=(k_mtm, D),
                seed=seed,
            )
        i += 1

    return samples


@nb.njit()
def compute_sum_subsets_norms(dists: np.ndarray) -> np.ndarray:
    r"""TODO: finish

    exploits the tree structure of the subsets.

    used in step 3 of ``compute_nlratio_prior_proposal``

    note: summing for all subset of neighbors, including the empty set, excluding the full set of neighbors

    Parameters
    ----------
    dists : np.ndarray of shape (N_neighbors, k_mtm, D)
        _description_

    Returns
    -------
    np.ndarray
        _description_
    """
    N_neighbors, k_mtm, D = dists.shape
    sums_ = np.zeros((2**N_neighbors - 1, k_mtm, D))

    if N_neighbors == 1:
        # sums_ : (1, k_mtm + 1, D)
        return sums_

    if N_neighbors == 2:
        # sums_ : (3, k_mtm + 1, D)
        sums_[1, :] = dists[0, :] * 1
        sums_[2, :] = dists[1, :] * 1
        return sums_

    if N_neighbors == 3:
        # sums_ : (7, k_mtm + 1, D)
        # 1st neighbor
        sums_[1, :] = dists[0, :] * 1
        sums_[2, :] = sums_[1, :] + dists[1, :]
        sums_[3, :] = sums_[1, :] + dists[2, :]
        # 2nd neighbor
        sums_[4, :] = dists[1, :] * 1
        sums_[5, :] = sums_[4, :] + dists[2, :]
        # 3rd neighbor
        sums_[6, :] = dists[2, :] * 1
        return sums_

    if N_neighbors == 4:
        # sums_ : (15, k_mtm + 1, D)
        # 1st neighbor
        sums_[1, :] = dists[0, :] * 1
        sums_[2, :] = sums_[1, :] + dists[1, :]
        sums_[3, :] = sums_[2, :] + dists[2, :]
        sums_[4, :] = sums_[2, :] + dists[3, :]
        sums_[5, :] = sums_[1, :] + dists[2, :]
        sums_[6, :] = sums_[5, :] + dists[3, :]
        sums_[7, :] = sums_[1, :] + dists[3, :]
        # 2nd neighbor
        sums_[8, :] = dists[1, :] * 1
        sums_[9, :] = sums_[8, :] + dists[2, :]
        sums_[10, :] = sums_[9, :] + dists[3, :]
        sums_[11, :] = sums_[8, :] + dists[3, :]
        # 3rd neighbor
        sums_[12, :] = dists[2, :] * 1
        sums_[13, :] = sums_[12, :] + dists[3, :]
        # 4th neighbor
        sums_[14, :] = dists[3, :] * 1
        return sums_

    else:
        assert np.min(sums_) > 0
        return sums_


@nb.njit(fastmath=True)
def numba_logsumexp_stable(x: np.ndarray) -> float:
    r"""stable numba implementation of the ``logsumexp`` function. Used in step 4 of ``compute_nlratio_prior_proposal``.

    Parameters
    ----------
    x : np.ndarray
        array of real values

    Returns
    -------
    float
        result of the ``logsumexp`` function
    """
    x_max = np.max(x)
    res = 0.0
    for j in range(x.size):
        res += np.exp(x[j] - x_max)
    res = np.log(res) + x_max
    return res


@nb.njit()
def compute_nlratio_prior_proposal(
    new_Theta: np.ndarray,
    spatial_list_edges: np.ndarray,
    spatial_weights: np.ndarray,
    idx_pix: np.ndarray,
    candidates_pix: np.ndarray,
) -> np.ndarray:
    """evaluates the negative log ratio of prior proposal in the MTM-chromatic Gibbs kernel in ``MySampler``

    Note
    ----
    The function works in four steps :

    * step 1: get the values of the neighbors of each pixel in ``idx_pix`` using ``get_neighboring_pixels``
    * step 2: compute the vector distances between the candidate and the pixel neighbors
    * step 3: compute list of sums of subsets using ``compute_sum_subsets_norms``
    * step 4: evaluate ``LogSumExp`` to obtain the log pdf of the Gaussian mixture, using the stable ``numba`` implementation.

    Parameters
    ----------
    new_Theta : np.ndarray
        iterate
    spatial_list_edges : np.ndarray
        list of edges, defined as pairs of pixel indices. Generated by the class:`SpatialPrior`.
    spatial_weights : np.ndarray
        regularization weights of the spatial prior
    idx_pix : np.ndarray of shape (n_pix,)
        array of indices of pixels for which candidates are to be proposed
    candidates_pix : np.ndarray of shape (n_pix, k_mtm, D)
        array of the candidate values

    Returns
    -------
    np.ndarray of shape (n_pix, k_mtm)
        negative log ratio of prior proposal
    """
    n_pix, k_mtm, D = candidates_pix.shape

    nl_ratio = np.zeros((n_pix, k_mtm))
    for i in range(n_pix):
        idx_1_pix = idx_pix[i]

        # * step 1: get neighbors
        neighbors = get_neighboring_pixels(
            new_Theta, spatial_list_edges, idx_1_pix
        )  # (N_neighbors, D)
        N_neighbors = neighbors.shape[0]

        # * step 2: compute distances between candidate and its neighbors
        dists = (
            np.expand_dims(candidates_pix[i], 0) - np.expand_dims(neighbors, 1)
        ) ** 2  # (N_neighbors, k_mtm, D)
        assert dists.shape == (N_neighbors, k_mtm, D)

        # * step 3: compute list of sums of subsets
        sums_norms = compute_sum_subsets_norms(
            dists
        )  # (2 ** N_neighbors - 1, k_mtm, D)
        assert sums_norms.shape == (2**N_neighbors - 1, k_mtm, D)

        # * step 4: LogSumExp
        inter_ = 2 * np.expand_dims(np.expand_dims(spatial_weights, 0), 0) * sums_norms
        # (2 ** N_neighbors - 1, k_mtm, D)
        for k in range(k_mtm):
            for d in range(D):
                nl_ratio[i, k] += numba_logsumexp_stable(inter_[:, k, d])

    return nl_ratio



import os
from typing import List, Union

import numpy as np

from beetroots.modelling.priors.spatial_prior_params import SpatialPriorParams
from beetroots.sampler.utils.my_sampler_params import MySamplerParams
from beetroots.simulations.abstract_simulation import Simulation
from beetroots.simulations.astro import data_validation
from beetroots.simulations.astro.forward_map.abstract_poly_reg import (
    SimulationPolynomialReg,
)
from beetroots.simulations.astro.observation.abstract_toy_case import SimulationToyCase
from beetroots.simulations.astro.posterior_type.abstract_direct import (
    SimulationMySampler,
)


class SimulationToyCaseNN(
    Simulation, SimulationToyCase, SimulationPolynomialReg, SimulationMySampler
):
    __slots__ = (
        "path_output_sim",
        "path_img",
        "path_raw",
        "path_data_csv",
        "path_data_csv_in",
        "path_data_csv_out",
        "path_data_csv_out_mcmc",
        "path_data_csv_out_optim_map",
        "path_data_csv_out_optim_mle",
        "N",
        "D",
        "D_no_kappa",
        "L",
        "list_names",
        "list_names_plot",
        "cloud_name",
        "max_workers",
        "list_lines_fit",
        "list_lines_valid",
        "Theta_true_scaled",
        "map_shaper",
        "plots_estimator",
    )

    def setup(
        self,
        forward_model_name: str,
        angle: float,
        #
        sigma_a_float: float,
        sigma_m_float: float,
        omega_float: float,
        #
        indicator_margin_scale: float,
        lower_bounds_lin: np.ndarray,
        upper_bounds_lin: np.ndarray,
        #
        with_spatial_prior: bool = True,
        spatial_prior_params: Union[None, SpatialPriorParams] = None,
        list_gaussian_approx_params: List[str] = [],
        list_mixing_model_params: List[str] = [],
    ):
        self.list_lines_valid = []

        scaler, forward_map = self.setup_forward_map(
            forward_model_name=forward_model_name,
            angle=angle,
        )

        sigma_a = sigma_a_float * np.ones((self.N, self.L))
        sigma_m = sigma_m_float * np.ones((self.N, self.L))
        omega = omega_float * np.ones((self.N, self.L))

        syn_map, y = self.setup_observation(
            scaler=scaler,
            forward_map=forward_map,
            sigma_a=sigma_a,
            sigma_m=sigma_m,
            omega=omega,
        )

        # run setup
        dict_posteriors, scaler, prior_indicator_1pix = self.setup_posteriors(
            scaler=scaler,
            forward_map=forward_map,
            y=y,
            sigma_a=sigma_a,
            sigma_m=sigma_m,
            omega=omega,
            syn_map=syn_map,
            with_spatial_prior=with_spatial_prior,
            spatial_prior_params=spatial_prior_params,
            indicator_margin_scale=indicator_margin_scale,
            lower_bounds_lin=lower_bounds_lin,
            upper_bounds_lin=upper_bounds_lin,
            list_gaussian_approx_params=list_gaussian_approx_params,
            list_mixing_model_params=list_mixing_model_params,
        )

        y_valid = None
        sigma_a_valid = None
        omega_valid = None

        return (
            dict_posteriors,
            scaler,
            prior_indicator_1pix,
            y_valid,
            sigma_a_valid,
            omega_valid,
        )


if __name__ == "__main__":
    to_run_optim_map = True
    to_run_mcmc = True

    N_1_side = 10

    sigma_a_float = 1.38715e-10
    sigma_m_float = np.log(1.1)
    omega_float = 3 * sigma_a_float

    indicator_margin_scale = 1e-1
    lower_bounds_lin = np.array([1e-1, 1e5, 1e0, 1e0])
    upper_bounds_lin = np.array([1e1, 1e9, 1e5, 4e1])

    if N_1_side == 10:
        my_sampler_params_mle = MySamplerParams(
            3e-3, 1e-5, 0.99, np.array([0.2, 0.8]), 250, False, False
        )  # 0.05
        my_sampler_params_map = MySamplerParams(
            1e-3, 1e-5, 0.99, np.array([0.2, 0.8]), 250, False, False
        )  # 0.2
        my_sampler_params_mcmc = MySamplerParams(
            5e-4, 1e-5, 0.99, np.array([0.5, 0.5]), 50, True, True
        )
        initial_regu_weights = np.array([1.0, 1.0, 1.0, 1.0])
    elif N_1_side == 30:
        my_sampler_params_mle = MySamplerParams(
            3e-3, 1e-5, 0.99, np.array([0.2, 0.8]), 250, False, False
        )  # 0.05
        my_sampler_params_map = MySamplerParams(
            3e-4, 1e-5, 0.99, np.array([0.5, 0.5]), 50, False, False
        )  # 0.2
        my_sampler_params_mcmc = MySamplerParams(
            1e-4, 1e-5, 0.99, np.array([0.5, 0.5]), 50, True, True
        )
        initial_regu_weights = np.array([10.0, 1.0, 2.0, 4.0])
    elif N_1_side == 64:
        my_sampler_params_mle = MySamplerParams(
            1e-3, 1e-5, 0.99, np.array([0.2, 0.8]), 250, False, False
        )
        my_sampler_params_map = MySamplerParams(
            5e-4, 1e-5, 0.99, np.array([0.5, 0.5]), 20, False, False
        )
        my_sampler_params_mcmc = MySamplerParams(
            3e-5, 1e-5, 0.99, np.array([0.5, 0.5]), 10, True, True
        )
        initial_regu_weights = np.array([10.0, 2.0, 3.0, 4.0])

    elif N_1_side == 90:
        my_sampler_params_mle = MySamplerParams(
            1e-3, 1e-5, 0.99, np.array([0.2, 0.8]), 250, False, False
        )
        my_sampler_params_map = MySamplerParams(
            5e-4, 1e-5, 0.99, np.array([0.5, 0.5]), 250, False, False
        )
        my_sampler_params_mcmc = MySamplerParams(
            3e-6, 1e-5, 0.99, np.array([0.5, 0.5]), 50, True, True
        )

        initial_regu_weights = np.array([10.0, 4.0, 5.0, 3.0])

    elif N_1_side == 100:
        my_sampler_params_mle = MySamplerParams(
            1e-3, 1e-5, 0.99, np.array([0.2, 0.8]), 250, False, False
        )
        my_sampler_params_map = MySamplerParams(
            5e-4, 1e-5, 0.99, np.array([0.5, 0.5]), 250, False, False
        )
        my_sampler_params_mcmc = MySamplerParams(
            1e-6, 1e-5, 0.99, np.array([0.5, 0.5]), 20, True, True
        )
        initial_regu_weights = np.array([10.0, 4.0, 5.0, 3.0])

    else:
        raise NotImplementedError("invalid N")

    # basic example for debugging purposes
    spatial_prior_params = SpatialPriorParams(
        name="L2-laplacian",
        use_next_nearest_neighbors=False,
        initial_regu_weights=initial_regu_weights,
    )

    list_lines_fit = [
        "co_v0_j4__v0_j3",
        "co_v0_j5__v0_j4",
        "co_v0_j6__v0_j5",
        "co_v0_j7__v0_j6",
        "co_v0_j8__v0_j7",
        "co_v0_j9__v0_j8",
        "co_v0_j10__v0_j9",
        "co_v0_j11__v0_j10",
        "co_v0_j12__v0_j11",
        "co_v0_j13__v0_j12",
    ]

    simulation = SimulationToyCaseNN(
        N_1_side**2,
        max_workers=30,
        list_lines_fit=list_lines_fit,
    )

    (
        dict_posteriors,
        scaler,
        prior_indicator_1pix,
        y_valid,  # None
        sigma_a_valid,  # None
        omega_valid,  # None
    ) = simulation.setup(
        forward_model_name="polynomial_regression_deg6",
        angle=0.0,
        #
        sigma_a_float=sigma_a_float,
        sigma_m_float=sigma_m_float,
        omega_float=omega_float,
        #
        indicator_margin_scale=indicator_margin_scale,
        lower_bounds_lin=lower_bounds_lin,
        upper_bounds_lin=upper_bounds_lin,
        #
        with_spatial_prior=True,
        spatial_prior_params=spatial_prior_params,
        #
        list_gaussian_approx_params=[],
        list_mixing_model_params=[
            {
                "path_transition_params": f"{os.path.dirname(os.path.abspath(__file__))}/../../../../../data/toycases/best_params.csv"
            }
        ],
    )
    simulation.save_and_plot_setup(
        dict_posteriors,
        lower_bounds_lin,
        upper_bounds_lin,
        scaler,
    )
    # * Optim MAP
    if to_run_optim_map:
        simulation.inversion_optim_map(
            dict_posteriors=dict_posteriors,
            scaler=scaler,
            my_sampler_params=my_sampler_params_map,
            N_MCMC=1,
            T_MC=200,
            T_BI=20,
            batch_size=20,
            freq_save=1,
            start_from=None,
        )

    # * MCMC
    if to_run_mcmc:
        simulation.inversion_mcmc(
            dict_posteriors=dict_posteriors,
            scaler=scaler,
            my_sampler_params=my_sampler_params_mcmc,
            N_MCMC=1,
            T_MC=1_000,
            T_BI=100,
            #
            plot_1D_chains=True,
            plot_2D_chains=True,
            plot_ESS=True,
            plot_comparisons_yspace=True,
            #
            batch_size=10,
            freq_save=1,
            start_from=None,
            #
            regu_spatial_N0=np.infty,
            regu_spatial_scale=1.0,
            regu_spatial_vmin=1e-8,
            regu_spatial_vmax=1e8,
        )



# import json
import os
import pickle
import warnings
from typing import Dict, Optional, Tuple

import numpy as np

from beetroots.modelling.forward_maps.neural_network_approx import NeuralNetworkApprox
from beetroots.simulations.astro.forward_map.abstract_forward_map import (
    SimulationForwardMap,
)
from beetroots.space_transform.transform import MyScaler

warnings.filterwarnings("ignore")


class SimulationNN(SimulationForwardMap):
    r"""abstract class for to set up the forward map to an already defined neural network for an inversion of astrophysical data"""

    def setup_forward_map(
        self,
        forward_model_name: str,
        force_use_cpu: bool,
        dict_fixed_params: Dict[str, Optional[float]],
        dict_is_log_scale_params: Dict[str, bool],
    ) -> Tuple[MyScaler, NeuralNetworkApprox]:
        print(f"fixed values: {dict_fixed_params}")

        with open(
            f"{self.MODELS_PATH}/{forward_model_name}/scaler.pickle",
            "rb",
        ) as file_:
            scaler_sklearn = pickle.load(file_)

        scaler = MyScaler(
            mean_=scaler_sklearn.mean_.flatten(),
            std_=scaler_sklearn.scale_.flatten(),
            list_is_log=list(dict_is_log_scale_params.values()),
        )

        # transformation from linear scale (in degrees) to scaled

        # angle_scaled = (angle - 30.0) / 20.0

        # eg {"kappa":None, "Pth":None, "G0":None, "AV":None, "angle":0.}
        dict_fixed_params_scaled = self.scale_dict_fixed_params(
            scaler, dict_fixed_params
        )

        # load forward model
        forward_map = NeuralNetworkApprox(
            forward_model_name,
            dict_fixed_params_scaled,
            device="cpu" if force_use_cpu else None,
        )
        forward_map.restrict_to_output_subset(self.list_lines_fit)
        return scaler, forward_map



import abc
import os

import numpy as np

# from beetroots.inversion.plots.map_shaper import MapShaper
from beetroots.inversion.plots.plots_2d_setup import Plots2DSetup
from beetroots.inversion.plots.plots_estimator import PlotsEstimator
from beetroots.space_transform.abstract_transform import Scaler


class SimulationObservation(abc.ABC):
    @abc.abstractmethod
    def setup_observation(self):
        pass

    def save_and_plot_setup(
        self,
        dict_posteriors: dict,
        lower_bounds_lin: np.ndarray,
        upper_bounds_lin: np.ndarray,
        scaler: Scaler,
    ):
        for model_name in list(dict_posteriors.keys()):
            folder_path = f"{self.path_raw}/{model_name}"
            if not os.path.isdir(folder_path):
                os.mkdir(folder_path)

        if self.N > 1:
            key = list(dict_posteriors.keys())[0]
            if isinstance(dict_posteriors[key], list):
                dict_sites_ = dict_posteriors[key][1].dict_sites
                y = dict_posteriors[key][0].likelihood.y * 1
                sigma_a = dict_posteriors[key][0].likelihood.sigma * 1
                omega = dict_posteriors[key][0].likelihood.omega * 1

            else:
                dict_sites_ = dict_posteriors[key].dict_sites
                y = dict_posteriors[key].likelihood.y * 1
                sigma_a = dict_posteriors[key].likelihood.sigma_a * 1
                omega = dict_posteriors[key].likelihood.omega * 1

            Plots2DSetup(
                self.path_img,
                self.map_shaper,
                self.N,
                self.pixels_of_interest,
            ).plot_all(
                y,
                sigma_a,
                omega,
                self.list_lines_fit,
                dict_sites_,
            )

            self.plots_estimator = PlotsEstimator(
                self.map_shaper,
                self.list_names_plots,
                lower_bounds_lin,
                upper_bounds_lin,
                self.list_idx_sampling,
                self.pixels_of_interest,
            )

            if self.Theta_true_scaled is not None:
                folder_path = f"{self.path_img}/estimators"
                folder_path_inter = f"{folder_path}/true"

                for path_ in [folder_path, folder_path_inter]:
                    if not os.path.isdir(path_):
                        os.mkdir(path_)

                Theta_true_scaled_full = np.zeros((self.N, self.D))
                Theta_true_scaled_full[
                    :, self.list_idx_sampling
                ] = self.Theta_true_scaled

                Theta_true_lin = scaler.from_scaled_to_lin(Theta_true_scaled_full)

                for d in range(self.D):
                    if self.list_fixed_values[d] is not None:
                        Theta_true_scaled_full[:, d] += self.list_fixed_values[d]

                self.plots_estimator.plot_estimator(
                    Theta_true_lin,
                    "true",
                    folder_path_inter,
                )

        else:
            self.plots_estimator = None
        return



from typing import Dict, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors
from scipy import interpolate


def plot_1D_hist(
    list_Theta_lin_seed: np.ndarray,
    n: Union[int, None],
    d: int,
    folder_path: str,
    title: str,
    lower_bounds_lin: Optional[np.ndarray],
    upper_bounds_lin: Optional[np.ndarray],
    seed: Optional[int] = None,
    estimator: Optional[float] = None,
    IC_low: Optional[float] = None,
    IC_high: Optional[float] = None,
    true_val: Optional[float] = None,
):
    if len(list_Theta_lin_seed.shape) == 3:
        list_Theta_lin_nd = list_Theta_lin_seed[:, n, d] * 1
    else:
        list_Theta_lin_nd = list_Theta_lin_seed * 1

    if lower_bounds_lin is None:
        lower_bounds_lin = np.min(list_Theta_lin_seed) * np.ones((10,))
    if upper_bounds_lin is None:
        upper_bounds_lin = np.max(list_Theta_lin_seed) * np.ones((10,))

    plt.figure(figsize=(8, 6))
    plt.title(title)

    assert lower_bounds_lin is not None
    if list_Theta_lin_seed.min() > 0 and lower_bounds_lin.min() > 0:
        plt.hist(np.log10(list_Theta_lin_nd), bins=100, label="samples")
    else:
        plt.hist(list_Theta_lin_nd, bins=100, label="samples")

    if estimator is not None:
        if list_Theta_lin_seed.min() > 0 and lower_bounds_lin.min() > 0:
            plt.axvline(np.log10(estimator), c="orange", ls="-", label="mean")
        else:
            plt.axvline(estimator, c="orange", ls="-", label="mean")

    if np.all([IC_low is not None, IC_high is not None]):
        assert IC_low is not None and IC_high is not None
        if list_Theta_lin_seed.min() > 0 and lower_bounds_lin.min() > 0:
            plt.axvline(np.log10(IC_low), c="k", ls="--", label="CI")
            plt.axvline(np.log10(IC_high), c="k", ls="--")
        else:
            plt.axvline(IC_low, c="k", ls="--", label="CI")
            plt.axvline(IC_high, c="k", ls="--")

    if true_val is not None:
        if list_Theta_lin_seed.min() > 0 and lower_bounds_lin.min() > 0:
            plt.axvline(np.log10(true_val), c="red", ls="--", label="true")
        else:
            plt.axvline(true_val, c="red", ls="--", label="true")

    plt.grid()
    plt.legend()
    # plt.tight_layout()
    if seed is not None:
        if n is not None:
            plt.savefig(f"{folder_path}/hist_n{n}_d{d}_seed{seed}.PNG")
        else:
            plt.savefig(f"{folder_path}/hist_d{d}_seed{seed}.PNG")
    else:
        if n is not None:
            plt.savefig(f"{folder_path}/hist_n{n}_d{d}_overall.PNG")
        else:
            plt.savefig(f"{folder_path}/hist_d{d}_overall.PNG")

    plt.close()
    return


def plot_1D_chain(
    list_Theta_lin_nd: np.ndarray,
    n: Optional[int],
    d: int,
    folder_path: str,
    title: str,
    lower_bounds_lin: Optional[np.ndarray],
    upper_bounds_lin: Optional[np.ndarray],
    N_MCMC: int,
    T_MC: int,
    T_BI: int,
    true_val: Optional[float] = None,
    is_u: bool = False,
) -> None:
    assert len(list_Theta_lin_nd.shape) == 1  # (N_MCMC * (T_MC - T_BI),)

    if lower_bounds_lin is None:
        lower_bounds_lin = np.min(list_Theta_lin_nd) * np.ones((10,))
    if upper_bounds_lin is None:
        upper_bounds_lin = np.max(list_Theta_lin_nd) * np.ones((10,))

    assert lower_bounds_lin is not None

    plt.figure(figsize=(8, 6))
    plt.title(title)
    plt.plot(list_Theta_lin_nd, label="MC chain")

    for seed in range(1, N_MCMC):
        if seed == 1:
            plt.axvline(
                seed * (T_MC - T_BI),
                c="k",
                ls="-",
                label="new MC",
            )
        else:
            plt.axvline(seed * (T_MC - T_BI) - seed, c="k", ls="-")

    if true_val is not None:
        plt.axhline(true_val, c="r", ls="--", label="true value")

    if list_Theta_lin_nd.min() > 0 and lower_bounds_lin.min() > 0:
        plt.yscale("log")

    plt.grid()
    plt.legend()
    # plt.tight_layout()

    if is_u:
        filename = f"{folder_path}/mc_1D_n{n}_ell{d}.PNG"
    else:
        filename = f"{folder_path}/mc_1D_n{n}_d{d}.PNG"
    plt.savefig(filename, bbox_inches="tight")
    plt.close()


def plot_2D_chain(
    list_Theta_lin_seed: np.ndarray,
    n: int,
    d1: int,
    d2: int,
    model_name: str,
    seed: int,
    folder_path: str,
    list_names: list,
    Theta_estimator: np.ndarray,
    Theta_true: Optional[np.ndarray] = None,
):
    # assert list_Theta_lin_seed.shape in [
    #     (self.T_MC - self.T_BI, self.N, self.D),
    #     (self.N_MCMC * (self.T_MC - self.T_BI), self.N, self.D),
    # ]
    freq = 5
    plt.figure(figsize=(8, 6))

    title = f"joint MC of ({list_names[d1]}, {list_names[d2]}) for pixel n={n}"
    plt.title(title)
    if list_Theta_lin_seed.min() > 0:
        plt.scatter(
            np.log10(list_Theta_lin_seed[::freq, n, d1]).flatten(),
            np.log10(list_Theta_lin_seed[::freq, n, d2]).flatten(),
            c=np.arange(list_Theta_lin_seed[::freq].shape[0]),
            s=10,
            label="samples",
        )
    else:
        plt.scatter(
            list_Theta_lin_seed[::freq, n, d1].flatten(),
            list_Theta_lin_seed[::freq, n, d2].flatten(),
            c=np.arange(list_Theta_lin_seed[::freq].shape[0]),
            s=10,
            label="samples",
        )
    if Theta_true is not None:
        plt.plot(
            np.log10([Theta_true[n, d1]]),
            np.log10([Theta_true[n, d2]]),
            "r+",
            ms=20,
            label="truth",
        )

    plt.colorbar()
    plt.legend()
    plt.grid()
    if list_Theta_lin_seed.min() > 0:
        plt.xlabel(r"$\log$ " + list_names[d1])
        plt.ylabel(r"$\log$ " + list_names[d2])
    else:
        plt.xlabel(list_names[d1])
        plt.ylabel(list_names[d2])
    # plt.tight_layout()
    plt.savefig(f"{folder_path}/mc_2D_n{n}_d1{d1}_d2{d2}_seed{seed}_chain.PNG")
    plt.close()


def plot_2D_hist(
    list_Theta_lin_seed: np.ndarray,
    n: int,
    d1: int,
    d2: int,
    model_name: str,
    folder_path: str,
    list_names: list,
    lower_bounds_lin: np.ndarray,
    upper_bounds_lin: np.ndarray,
    Theta_MMSE: np.ndarray,
    true_val: Optional[np.ndarray] = None,
    seed: Union[int, str] = "overall",
    point_challenger: Dict = {},
):
    # assert list_Theta_lin_seed.shape in [
    #     (self.T_MC - self.T_BI, self.N, self.D),
    #     (self.N_MCMC * (self.T_MC - self.T_BI), self.N, self.D),
    # ]
    def _set_edges():
        list_Theta_lin_seed_used = list_Theta_lin_seed * 1

        # x-axis
        if list_Theta_lin_seed[:, 0].min() > 0 and lower_bounds_lin[d1] > 0:
            bounds_theta = (
                np.log10(lower_bounds_lin[d1]) - 0.1,
                np.log10(upper_bounds_lin[d1]) + 0.1,
            )
            list_Theta_lin_seed_used[:, 0] = np.log10(list_Theta_lin_seed_used[:, 0])
            is_theta_log = True
        else:
            bounds_theta = (
                lower_bounds_lin[d1] - 0.1,
                upper_bounds_lin[d1] + 0.1,
            )
            is_theta_log = False

        # y-axis
        if list_Theta_lin_seed[:, 1].min() > 0 and lower_bounds_lin[d2] > 0:
            bounds_y = (
                np.log10(lower_bounds_lin[d2]) - 0.1,
                np.log10(upper_bounds_lin[d2]) + 0.1,
            )
            list_Theta_lin_seed_used[:, 1] = np.log10(list_Theta_lin_seed_used[:, 1])
            is_y_log = True
        else:
            bounds_y = (
                lower_bounds_lin[d2] - 0.1,
                upper_bounds_lin[d2] + 0.1,
            )
            is_y_log = False

        return list_Theta_lin_seed_used, bounds_theta, bounds_y, is_theta_log, is_y_log

    (
        list_Theta_lin_seed_used,
        bounds_theta,
        bounds_y,
        is_theta_log,
        is_y_log,
    ) = _set_edges()

    plt.figure(figsize=(8, 6))

    title = f"joint MC of ({list_names[d1]}, {list_names[d2]}) for pixel n={n}"
    plt.title(title)

    plt.hist2d(
        list_Theta_lin_seed_used[:, 0].flatten(),
        list_Theta_lin_seed_used[:, 1].flatten(),
        (100, 100),
        range=[[bounds_theta[0], bounds_theta[1]], [bounds_y[0], bounds_y[1]]],
        norm=colors.LogNorm(),
    )

    plt.plot(
        np.log10([Theta_MMSE[0]]) if is_theta_log else [Theta_MMSE[0]],
        np.log10([Theta_MMSE[1]]) if is_y_log else [Theta_MMSE[1]],
        "rx",
        ms=20,
        label="MMSE",
        markeredgewidth=3,
    )

    if len(point_challenger) > 0:
        x_chal = point_challenger["value"] * 1
        plt.plot(
            np.log10([x_chal[n, d1]]) if is_theta_log else [x_chal[n, d1]],
            np.log10([x_chal[n, d2]]) if is_y_log else [x_chal[n, d2]],
            "k+",
            ms=20,
            label=point_challenger["name"],
            markeredgewidth=3,
        )

    if true_val is not None:
        plt.plot(
            np.log10([true_val[0]]) if is_theta_log else [true_val[0]],
            np.log10([true_val[1]]) if is_y_log else [true_val[1]],
            "r+",
            ms=20,
            label="truth",
            markeredgewidth=3,
        )

    plt.legend(loc="best")

    # in case no point in acceptable set
    plt.colorbar()

    plt.grid()
    plt.xlabel(r"$\log$ " + list_names[d1] if is_theta_log else list_names[d1])
    plt.ylabel(r"$\log$ " + list_names[d2] if is_y_log else list_names[d2])
    # plt.tight_layout()

    if seed == "overall":
        filename = f"{folder_path}/hist2D_n{n}_d1{d1}_d2{d2}_overall_chain.PNG"
    else:
        filename = f"{folder_path}/hist2D_n{n}_d1{d1}_d2{d2}_seed{seed}"
        filename += "_chain.PNG"
    plt.savefig(filename)
    plt.close()


def plot_2D_proba_contours(
    list_Theta_lin_seed: np.ndarray,
    n: int,
    d1: int,
    d2: int,
    model_name: str,
    folder_path: str,
    list_names: list,
    lower_bounds_lin: np.ndarray,
    upper_bounds_lin: np.ndarray,
    Theta_MMSE: np.ndarray,
    true_val: Optional[np.ndarray] = None,
    seed: Union[int, str] = "overall",
    point_challenger: Dict = {},
):
    """plots the 2D contours of the 68% and 95% high probability regions

    inspired from https://stackoverflow.com/questions/37890550/python-plotting-percentile-contour-lines-of-a-probability-distribution
    """
    # assert list_Theta_lin_seed.shape in [
    #     (self.T_MC - self.T_BI, self.N, self.D),
    #     (self.N_MCMC * (self.T_MC - self.T_BI), self.N, self.D),
    # ]
    def _set_edges_and_hist(n_per_axis: int):
        list_Theta_lin_seed_used = list_Theta_lin_seed * 1

        # x-axis
        if list_Theta_lin_seed[:, 0].min() > 0 and lower_bounds_lin[d1] > 0:
            x_edges = np.linspace(
                np.log10(lower_bounds_lin[d1]) - 0.1,
                np.log10(upper_bounds_lin[d1]) + 0.1,
                n_per_axis,
            )
            list_Theta_lin_seed_used[:, 0] = np.log10(list_Theta_lin_seed_used[:, 0])
            is_theta_log = True
        else:
            x_edges = np.linspace(
                lower_bounds_lin[d1] - 0.1,
                upper_bounds_lin[d1] + 0.1,
                n_per_axis,
            )
            is_theta_log = False

        # y-axis
        if list_Theta_lin_seed[:, 1].min() > 0 and lower_bounds_lin[d2] > 0:
            y_edges = np.linspace(
                np.log10(lower_bounds_lin[d2]) - 0.1,
                np.log10(upper_bounds_lin[d2]) + 0.1,
                n_per_axis,
            )
            list_Theta_lin_seed_used[:, 1] = np.log10(list_Theta_lin_seed_used[:, 1])
            is_y_log = True
        else:
            y_edges = np.linspace(
                lower_bounds_lin[d2] - 0.1,
                upper_bounds_lin[d2] + 0.1,
                n_per_axis,
            )
            is_y_log = False

        H, x_edges, y_edges = np.histogram2d(
            list_Theta_lin_seed_used[:, 0].flatten(),
            list_Theta_lin_seed_used[:, 1].flatten(),
            bins=(x_edges, y_edges),
        )
        return H, x_edges, y_edges, is_theta_log, is_y_log

    n_per_axis = 100

    #! Contour levels must be increasing, ie percentiles must be decreasing
    percentiles_arr = np.array([0.95, 0.68])

    H, x_edges, y_edges, is_theta_log, is_y_log = _set_edges_and_hist(n_per_axis)
    H /= H.sum()
    t = np.linspace(0, H.max(), 1_000)
    integral = ((H >= t[:, None, None]) * H).sum(axis=(1, 2))

    # an error comes if the histogram grid is not fine enough to compute
    # the contours
    # Hence we progressively increase the grid size per axis until the grid
    # gets fine enough.
    # this error occurs frequently when the Markov chain size is small and
    # when the high proba region is very small compared to the acceptable
    # solutions set.
    while (integral[-1] > percentiles_arr[-1]) and (n_per_axis < 5e2):
        n_per_axis *= 2  # refine histogram grid
        H, x_edges, y_edges, _, _ = _set_edges_and_hist(n_per_axis)
        H /= H.sum()
        t = np.linspace(0, H.max(), 1_000)
        integral = ((H >= t[:, None, None]) * H).sum(axis=(1, 2))

    f = interpolate.interp1d(integral, t)

    if integral[-1] > percentiles_arr[-1]:
        percentiles_arr = percentiles_arr[integral[-1] < percentiles_arr]

        msg = "Issue with proba contour plot for (n, d1, d2) = "
        msg += f"({n}, {d1}, {d2})"
        print(msg)

        if percentiles_arr.size == 0:
            return

    # define contours
    t_contours = f(percentiles_arr)

    plt.figure(figsize=(8, 6))

    title = f"High Probability Region of ({list_names[d1]}, {list_names[d2]})"
    title += f" for pixel n={n}"
    plt.title(title)

    for percentile, t_contour in zip(percentiles_arr, t_contours):
        plt.contour(
            H.T,
            t_contours,
            extent=[x_edges[0], x_edges[-1], y_edges[0], y_edges[-1]],
            label=f"{100 * percentile} %",
        )

    if list_Theta_lin_seed.min() > 0 and lower_bounds_lin.min() > 0:
        plt.plot(
            np.log10([Theta_MMSE[0]]),
            np.log10([Theta_MMSE[1]]),
            "rx",
            ms=20,
            label="MMSE",
            markeredgewidth=3,
        )
    else:
        plt.plot(
            [Theta_MMSE[0]],
            [Theta_MMSE[1]],
            "rx",
            ms=20,
            label="MMSE",
            markeredgewidth=3,
        )

    if len(point_challenger) > 0:
        x_challenger = point_challenger["value"]
        if list_Theta_lin_seed.min() > 0 and lower_bounds_lin.min() > 0:
            plt.plot(
                np.log10([x_challenger[n, d1]]),
                np.log10([x_challenger[n, d2]]),
                "k+",
                ms=20,
                label=point_challenger["name"],
                markeredgewidth=3,
            )
        else:
            plt.plot(
                [x_challenger[n, d1]],
                [x_challenger[n, d2]],
                "k+",
                ms=20,
                label=point_challenger["name"],
                markeredgewidth=3,
            )

    if true_val is not None:
        if list_Theta_lin_seed.min() > 0 and lower_bounds_lin.min() > 0:
            plt.plot(
                np.log10([true_val[0]]),
                np.log10([true_val[1]]),
                "r+",
                ms=20,
                label="truth",
                markeredgewidth=3,
            )
        else:
            plt.plot(
                [true_val[0]],
                [true_val[1]],
                "r+",
                ms=20,
                label="truth",
                markeredgewidth=3,
            )

    plt.legend(loc="best")

    # in case no point in acceptable set
    # plt.colorbar()

    plt.grid()
    plt.xlabel(r"$\log$ " + list_names[d1] if is_theta_log else list_names[d1])
    plt.ylabel(r"$\log$ " + list_names[d2] if is_y_log else list_names[d2])
    # plt.tight_layout()

    if seed == "overall":
        filename = f"{folder_path}/HPR_2D_n{n}_d1{d1}_d2{d2}_overall_chain.PNG"
    else:
        filename = f"{folder_path}/HPR_2D_n{n}_d1{d1}_d2{d2}_seed{seed}"
        filename += "_chain.PNG"
    plt.savefig(filename)
    plt.close()
    return



import abc
from typing import List


class ResultsUtil(abc.ABC):
    """abstract class for the extraction of a specific result from the data saved during an inversion. All the daughter classes have three key methods:

    * ``read_data``, to read the data necessary for the computation of the specific result of interest
    * ``create_folders``, to create the folders where the specific result is to be saved
    * ``main``, to run the two previous methods and compute the specific result
    """

    @abc.abstractmethod
    def read_data(self, list_chains_folders: List[str], **kwargs):
        r"""read the data necessary for the computation of the specific result of interest

        Parameters
        ----------
        list_chains_folders : List[int]
            list of the paths to the folders containing the raw results
        """
        pass

    @abc.abstractmethod
    def create_folders(self, **kwargs) -> str:
        r"""create the folder where the specific result is to be saved

        Returns
        -------
        str
            path to the folder where the specific result is to be saved
        """
        pass

    @abc.abstractmethod
    def main(self, **kwargs):
        r"""runs the two previous methods and compute the specific result"""
        pass
